/*******************************************************************************
**	ATT20c505 / ATT20c504 ramdac driver
********************************************************************************
**
   Copyright (C) 1995-1997 by
**	Steffen Seeger,	[seeger@ggi-project.org]
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger

#ifndef	DRIVER_NAME
	#define	DRIVER_NAME	"AT&T 20C505/20C504 DAC driver"
#endif
#define	DRIVER_REV	"$Revison:$"

#define	DEBUG_LEVEL	0	/* do silent checking */

#undef	DEBUG

#define DAC_PRECISION	8	/* do 8 bit R,G,B resolution */

#include <kgi/module.h>
#include <kgi/types.h>
#include <kgi/kgi_commands.h>
#include "ATT/20c505.h"


#define PCLK DAC_C2_PCLK0     /* or DAC_C2_PCLK1 ? (doesn't matter on 20SV) */

enum RAMDAC_types {

	ATT20C504_85	= 0,
	ATT20C504_11	= 1,

	ATT20C505_11	= 2,
	ATT20C505_13	= 3,
	ATT20C505_17	= 4,

	BT485_11	= 5,
	BT485_13	= 6,

	DAC_UNKNOWN	= -1
};

#define ATT20C504 ATT20C504_85
#define ATT20C505 ATT20C505_11
#define	BT485 BT485_11

#define	SUPPORTED (GM_TEXT16 | GM_1BIT | GM_4BIT | GM_8BIT \
		  | GM_15BIT | GM_16BIT | GM_32BIT)

static struct kgi_ramdac att_params[7] =
{
       {
		"AT&T Microelectronics",	/* manufacturer		*/
		 "ATT 20C504-85",		/* model		*/
		 SUPPORTED,			/* supported modes	*/
		 FL_DAC_8BIT | FL_DAC_GAMMA,	/* flags		*/
		 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		 DAC_8BIT_PB | 			/* device type		*/
		 DAC_32BIT_PB | DAC_VRAM,
		  {       0,  85000000 },	/* normal mode		*/
		  {       0,         0 },	/* 2xclk mode		*/

		 {{	  0,  75000000 },	/* 	KGIGT_TEXT		*/
		  {       0,         0 },	/*	KGIGT_TEXT32	*/
		  {       0,  85000000 },	/* 	KGIGT_1BIT		*/
		  {	  0,  85000000 },	/* 	KGIGT_4BIT		*/
		  {       0,  85000000 },	/* 	KGIGT_8BIT		*/	
		  {	  0,  85000000 },	/* 	KGIGT_15BIT	*/
		  { 	  0,  85000000 },	/* 	KGIGT_16BIT	*/
		  {    	  0,         0 },	/* 	KGIGT_24BIT	*/
	 	 {	  0,  75000000 }}	/* 	KGIGT_32BIT	*/
	},

	{
		"AT&T Microelectronics",	/* manufacturer		*/
		"ATT 20C504-11",		/* model		*/
		 SUPPORTED,			/* supported modes	*/
		 FL_DAC_8BIT | FL_DAC_GAMMA,	/* flags		*/
		 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		 DAC_8BIT_PB | 			/* device type		*/
		 DAC_32BIT_PB | DAC_VRAM,
		  {       0, 110000000 },	/* limit in normal mode	*/
		  {       0,         0 },	/* limit in 2xclk mode	*/
 
		 {{	  0,  75000000 },	/* 	KGIGT_TEXT		*/
		  {       0,         0 },	/*	KGIGT_TEXT32	*/
		  {       0, 110000000 },	/* 	KGIGT_1BIT		*/
		  {	  0, 110000000 },	/* 	KGIGT_4BIT		*/
		  {       0, 110000000 },	/* 	KGIGT_8BIT		*/	
		  {	  0, 110000000 },	/* 	KGIGT_15BIT	*/
		  { 	  0, 110000000 },	/* 	KGIGT_16BIT	*/
		  {       0,         0 },	/* 	KGIGT_24BIT	*/
		  {	  0,  75000000 }}	/* 	KGIGT_32BIT	*/
	},

	{
		"AT&T Microelectronics",	/* manufacturer		*/
		"ATT 20C505-11",		/* model		*/
		SUPPORTED,			/* supported modes	*/
		FL_DAC_8BIT | FL_DAC_GAMMA |	/* flags		*/
		FL_DAC_2XCLOCK,
		{ 0x0FFF, 0x0FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		DAC_8BIT_PB | 			/* device type		*/
		DAC_32BIT_PB | DAC_VRAM,
		{       0, 110000000 },		/* normal mode		*/
		{45000000,  55000000 },		/* limit in 2xclk mode	*/

		{{	  0,  75000000 },	/* 	KGIGT_TEXT		*/
		 {        0,         0 },	/*	KGIGT_TEXT32	*/
		 {        0, 110000000 },	/* 	KGIGT_1BIT		*/
		 {	  0, 110000000 },	/* 	KGIGT_4BIT		*/
		 {        0, 110000000 },	/* 	KGIGT_8BIT		*/	
		 {	  0, 110000000 },	/* 	KGIGT_15BIT	*/
		 { 	  0, 110000000 },	/* 	KGIGT_16BIT	*/
		 {        0,         0 },	/* 	KGIGT_24BIT	*/
		 {	  0,  75000000 }}	/* 	KGIGT_32BIT	*/
	},

       {
		"AT&T Microelectronics",	/* manufacturer		*/
		 "ATT 20C505-13",		/* model		*/
		 SUPPORTED,			/* supported modes	*/
		 FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
		 FL_DAC_2XCLOCK,
		 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		 DAC_8BIT_PB | 			/* device type		*/
		 DAC_32BIT_PB | DAC_VRAM,
		  {       0, 110000000 },	/* normal mode		*/
		  {45000000,  67500000 },	/* 2xclk mode		*/

		 {{	  0,  85000000 },	/* 	KGIGT_TEXT		*/
		  {       0,         0 },	/*	KGIGT_TEXT32	*/
		  {       0, 135000000 },	/* 	KGIGT_1BIT		*/
		  {	  0, 135000000 },	/* 	KGIGT_4BIT		*/
		  {       0, 135000000 },	/* 	KGIGT_8BIT		*/	
		  {	  0, 135000000 },	/* 	KGIGT_15BIT	*/
		  { 	  0, 135000000 },	/* 	KGIGT_16BIT	*/
		  {       0,         0 },	/* 	KGIGT_24BIT	*/
		  {	  0,  85000000 }}	/* 	KGIGT_32BIT	*/
	},

	{
		"AT&T Microelectronics",	/* manufacturer		*/
		 "ATT 20C505-17",		/* model		*/
		 SUPPORTED,			/* supported modes	*/
		 FL_DAC_8BIT | FL_DAC_GAMMA |	/* flags		*/
		 FL_DAC_2XCLOCK,
		 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		 DAC_8BIT_PB | 			/* device type		*/
		 DAC_32BIT_PB | DAC_VRAM,
		  { 	  0, 110000000 },	/* normal mode		*/
		  {45000000,  85000000 },	/* 2xclk mode		*/

		 {{	  0,  85000000 },	/* 	KGIGT_TEXT		*/
		  {       0,         0 },	/*	KGIGT_TEXT32	*/
		  {       0, 135000000 },	/* 	KGIGT_1BIT		*/
		  {	  0, 175000000 },	/* 	KGIGT_4BIT		*/
		  {       0, 175000000 },	/* 	KGIGT_8BIT		*/	
		  {	  0, 175000000 },	/* 	KGIGT_15BIT	*/
		  { 	  0, 175000000 },	/* 	KGIGT_16BIT	*/
		  {       0,         0 },	/* 	KGIGT_24BIT	*/
		  {	  0,  85000000 }}	/* 	KGIGT_32BIT	*/
	},

	{
		"Brooktree Corporation",	/* manufacturer		*/
		"Bt 485-11",			/* model		*/
		 SUPPORTED & ~GM_1BIT,		/* supported modes	*/
		 FL_DAC_8BIT | FL_DAC_GAMMA |	/* flags		*/
		 FL_DAC_2XCLOCK,
		 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		 DAC_8BIT_PB | 			/* device type		*/
		 DAC_32BIT_PB | DAC_VRAM,
		  {       0, 110000000 },	/* normal mode		*/
		  {45000000,  55000000 },	/* limit in 2xclk mode	*/

		 {{	  0,  75000000 },	/* 	KGIGT_TEXT		*/
		  {       0,         0 },	/*	KGIGT_TEXT32	*/
		  {       0, 110000000 },	/* 	KGIGT_1BIT		*/
		  {	  0, 110000000 },	/* 	KGIGT_4BIT		*/
		  {       0, 110000000 },	/* 	KGIGT_8BIT		*/	
		  {	  0, 110000000 },	/* 	KGIGT_15BIT	*/
		  { 	  0, 110000000 },	/* 	KGIGT_16BIT	*/
		  {       0,         0 },	/* 	KGIGT_24BIT	*/
		  {	  0,  75000000 }}	/* 	KGIGT_32BIT	*/
	},

	{
		 "Brooktree Corporation",	/* manufacturer		*/
		 "Bt 485-13",			/* model		*/
		 SUPPORTED & ~GM_1BIT,		/* supported modes	*/
		 FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
		 FL_DAC_2XCLOCK,
		 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		 DAC_8BIT_PB | 			/* device type		*/
		 DAC_32BIT_PB | DAC_VRAM,
		  {       0, 110000000 },	/* normal mode		*/
		  {45000000,  67500000 },	/* 2xclk mode		*/

		 {{	  0,  85000000 },	/* 	KGIGT_TEXT		*/
		  {       0,         0 },	/*	KGIGT_TEXT32	*/
		  {       0, 135000000 },	/* 	KGIGT_1BIT		*/
		  {	  0, 135000000 },	/* 	KGIGT_4BIT		*/
		  {       0, 135000000 },	/* 	KGIGT_8BIT		*/	
		  {	  0, 135000000 },	/* 	KGIGT_15BIT	*/
		  { 	  0, 135000000 },	/* 	KGIGT_16BIT	*/
		  {       0,         0 },	/* 	KGIGT_24BIT	*/
		  {	  0,  85000000 }}	/* 	KGIGT_32BIT	*/
	}
};

/*
**	driver state info
*/

struct dac_private 
{
	kgi_uint	cur_hotx, cur_hoty;	/* cursor hot spot	*/
	kgi_uint	cur_x, cur_y;		/* current position	*/

	kgiu8	ctrl0, ctrl1, ctrl2, ctrl3;	/* control registers	*/
};

#define	CTRL0		private->ctrl0
#define	CTRL1		private->ctrl1
#define	CTRL2		private->ctrl2
#define	CTRL3		private->ctrl3
#define	CUR_HOTX	private->cur_hotx
#define	CUR_HOTY	private->cur_hoty
#define	CUR_X		private->cur_x
#define	CUR_Y		private->cur_y

#define	GET_PRIVATE	struct dac_private *private = KGIM_PRIVATE(dpy,ramdac)


/*
**	helper functions
*/

#include "IBM/clut.inc"		/* use dac_set_clut() from normal VGA */


static void dac_set_pointer_shape(struct kgi_display *dpy, 
	struct kgi_pointer64x64 *ptr)
{
	GET_PRIVATE;
	kgi_uint i;

	/*	update private state
	*/
	CTRL2 &= ~DAC_C2_CURSOR_TYPE;
	switch (ptr->mode) {

	case pmWindows:
		CTRL2 |= DAC_C2_WIN_CURSOR;
		break;

	case pmX11:
		CTRL2 |= DAC_C2_X11_CURSOR;
		break;

	case pmThreeColor:
		CTRL2 |= DAC_C2_3CO_CURSOR;
		break;

	default:
		INTERNAL_ERROR;
	}
	CUR_HOTX = ptr->hotx;
	CUR_HOTY = ptr->hoty;
	DAC_out8(dpy, CTRL2, DAC_CONTROL2);

	DEBUG2("set_pointer_shape: hot spot (%i,%i), ctrl3 %i", \
		CUR_HOTX, CUR_HOTY, CTRL3);

	/*	set pattern
	*/
	if (CTRL3 & DAC_C3_LARGE_CURSOR) {

		ASSERT((CTRL3 & DAC_C3_ADDR) == 0);
		NOTICE2("setting 64x64 cursor");

		DAC_out8(dpy, CTRL0 | DAC_C0_EXT_UNLOCK, DAC_CONTROL0);
		DAC_out8(dpy, DAC_PW_CRTL3_INDEX, DAC_PW_INDEX);
		DAC_out8(dpy, CTRL3, DAC_STATUSREG);
		DAC_out8(dpy, CTRL0, DAC_CONTROL0);

		DAC_out8(dpy, 0, DAC_PW_INDEX);
		DAC_outs8(dpy, DAC_CURSOR, ptr->xor_mask, 512);
		DAC_outs8(dpy, DAC_CURSOR, ptr->and_mask, 512);

	} else {

		kgiu8	*data;

		NOTICE2("setting 32x32 cursor");

		DAC_out8(dpy, 0, DAC_PW_INDEX);

		data = ptr->xor_mask;
		for (i = 33; --i; ) {

			DAC_outs8(dpy, DAC_CURSOR, data, 4);
			data += 8;
		}

		data = ptr->and_mask;
		for (i = 33; --i; ) {

			DAC_outs8(dpy, DAC_CURSOR, data, 4);
			data += 8;
		}
	}

	/*	set colors
	*/
	DAC_out8(dpy, 1, DAC_CW_INDEX);
	for (i = 0; i < 3; i++) {	

		DAC_out8(dpy, ptr->col[i].r >> SHIFT, DAC_C_DATA);
		DAC_out8(dpy, ptr->col[i].g >> SHIFT, DAC_C_DATA);
		DAC_out8(dpy, ptr->col[i].b >> SHIFT, DAC_C_DATA);
	}
}


static void dac_show_pointer(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	GET_PRIVATE;

	if ((CUR_X != x) || (CUR_Y != y)) {

		CUR_X = x;
		x += ((CTRL3 & DAC_C3_LARGE_CURSOR) ? 64 : 32) - 
			CUR_HOTX + dpy->mode.x.total - dpy->mode.x.blankend;

		CUR_Y = y;
		y += ((CTRL3 & DAC_C3_LARGE_CURSOR) ? 64 : 32) -
			CUR_HOTY + dpy->mode.y.total - dpy->mode.y.blankend;

		DAC_out8(dpy, x,      DAC_CURSOR_XL);
		DAC_out8(dpy, x >> 8, DAC_CURSOR_XH);
		DAC_out8(dpy, y,      DAC_CURSOR_YL);
		DAC_out8(dpy, y >> 8, DAC_CURSOR_YH);
	}
}

static void dac_hide_pointer(struct kgi_display *dpy)
{
	GET_PRIVATE;

	CUR_X = 0x0FFF;
	CUR_Y = 0x0FFF;

	DAC_out8(dpy, 0xFF, DAC_CURSOR_XL);
	DAC_out8(dpy, 0x0F, DAC_CURSOR_XH);
	DAC_out8(dpy, 0xFF, DAC_CURSOR_YL);
	DAC_out8(dpy, 0x0F, DAC_CURSOR_YH);
}


/*
**	KGI interface
*/

int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	kgiu8 status;
	kgi_uint model;
	struct dac_private *private;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	if (!(private = kgi_malloc_small(sizeof(*private)))) {

		ERROR("failed to allocate private data!");
		return -ENOMEM;
	}
	KGIM_PRIVATE(dpy,ramdac) = private;
	memset(private, 0, sizeof(*private));

	status = DAC_in8(dpy, DAC_STATUSREG);
	switch(status & DAC_ST_ID_MASK) {

	case DAC_BT485:
		CTRL3 |= DAC_C3_LARGE_CURSOR;
		model = BT485;
		break;

	case DAC_ATT20C504:
		model = ATT20C504;
		break;

	case DAC_ATT20C505:
		CTRL3 |= DAC_C3_LARGE_CURSOR;
		model = ATT20C505;
		break;

	case DAC_ATT20C506:
		kgi_free_small(private);
		KGIM_PRIVATE(dpy, ramdac) = NULL;
		ERROR("don't know how to handle ATT20C506 device.");
		return -E(RAMDAC, UNKNOWN);

	default:
		kgi_free_small(private);
		KGIM_PRIVATE(dpy, ramdac) = NULL;
		DEBUG0("unknown chip ID (%.2x).", status);
		return -E(RAMDAC, UNKNOWN);
	}

	#if DEBUG_LEVEL < 3

		if (CTRL3 & DAC_C3_LARGE_CURSOR) {

			NOTICE1("  .. using 64x64 cursor.");

		} else {

			NOTICE1("  .. using 32x32 cursor.");
		}
	#endif

	/*	Now enter the DAC parameters/functions into the kgi_card and
	**	kgi_display structs.
	*/
	card->ramdac = att_params + model;
	DEBUG1("  .. %s %s detected.", card->ramdac->manufact, \
		card->ramdac->model);
	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	kgi_free_small(KGIM_PRIVATE(dpy,ramdac));
	KGIM_PRIVATE(dpy,ramdac) = NULL;
	NOTICE(DRIVER_NAME " removed.");
}

int kgim_ramdac_check_mode (struct kgi_display *dpy, struct kgi_mode *mode, 
			enum kgi_tm_cmd *cmd)
{
	int div;
	int newclock = mode->dclk;
	struct kgi_ramdac *dac = dpy->card->ramdac;

	switch(mode->request.graphtype) {

	case KGIGT_TEXT16:	div =  1;	break;
	case KGIGT_1BIT:	div = 32;	break;
	case KGIGT_4BIT:	div =  8;	break;
	case KGIGT_8BIT:	div =  4;	break;
 
	case KGIGT_15BIT:
	case KGIGT_16BIT:	div =  2;	break;
  
	case KGIGT_32BIT:	div =  1;	break;
  
	case KGIGT_24BIT:	return -E(RAMDAC, UNKNOWN);

	default:
		DEBUG2("could not handle mode %x", mode->request.graphtype);
		return -E(RAMDAC, UNKNOWN);
	}

	switch (*cmd) {

	case CMD_PROPOSE:

		mode->lclk.mul = mode->clk.mul = mode->clk.div = 1;
	        mode->lclk.div = div;
		DEBUG2("%i:%i multiplex mode", mode->lclk.div, mode->lclk.mul);

		mode->flags &= ~FL_DAC_2XCLOCK;
		mode->flags |= FL_DAC_8BIT;
		if ((mode->request.graphtype == KGIGT_32BIT) || 
			(mode->request.graphtype == KGIGT_15BIT) ||
			(mode->request.graphtype == KGIGT_16BIT)) {

			mode->flags |= FL_DAC_GAMMA;
		}
		return EOK;

	case CMD_LOWER:

		if ((newclock < dac->clock[mode->request.graphtype].min) ||
			(newclock < dac->normal.min)) {

			DEBUG2("%i Hz DCLK is too low", newclock);
			return -E(RAMDAC, CLK_LOW_LIMIT);
		}

		if (dac->clock[mode->request.graphtype].max < newclock) {

    			newclock = dac->clock[mode->request.graphtype].max;
		}

		if (dac->normal.max < newclock) {

			NOTICE2("double clock mode not implemented");
			newclock = dac->normal.max;
		}

		DEBUG2("dclk = %i", newclock);
		mode->dclk   = newclock;
		return EOK;

	case CMD_RAISE:

		if ((newclock > dac->clock[mode->request.graphtype].max) ||
			(newclock > dac->normal.max)) {

			DEBUG2("%i Hz DCLK is too high", newclock);
			return -E(RAMDAC, CLK_HIGH_LIMIT);
		}

		if (newclock < dac->clock[mode->request.graphtype].min) {

			newclock = dac->clock[mode->request.graphtype].min;
		}

		if (newclock < dac->normal.min) {

			newclock = dac->normal.min;
		}

		mode->dclk = newclock;
		return EOK;

     case CMD_CHECK:
		if ((mode->lclk.mul != 1) || (mode->lclk.div != div) ||
			(mode->clk.mul != 1) || (mode->clk.div != 1)) {

			return -E(RAMDAC, UNKNOWN);
		}

		if ((mode->dclk < dac->normal.min) ||
			(mode->dclk <dac->clock[mode->request.graphtype].min)) {

			return -E(RAMDAC, CLK_LOW_LIMIT);
		}

		if ((mode->dclk > dac->normal.max) ||
			(mode->dclk >dac->clock[mode->request.graphtype].max)) {

			return -E(RAMDAC, CLK_HIGH_LIMIT);
		}

		return EOK;

	default:

		INTERNAL_ERROR;
		return -E(RAMDAC, UNKNOWN);
	}
}

void kgim_ramdac_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	GET_PRIVATE;
	enum kgi_sync synctype = dpy->card->monitor->synctype;

	/*	setup ctrl0
	*/
	CTRL0 = DAC_C0_8BIT;
	if (synctype & SYNC_ON_RED) {

		CTRL0 |= DAC_C0_SYNC_ON_RED;
	}
	if (synctype & SYNC_ON_GREEN) {

		CTRL0 |= DAC_C0_SYNC_ON_GREEN;
	}
	if (synctype & SYNC_ON_BLUE) {

		CTRL0 |= DAC_C0_SYNC_ON_BLUE;
	}
	if (synctype & SYNC_PEDESTRAL) {

		CTRL0 |= DAC_C0_BLANK_LEVEL;
	}


	/*	setup ctrl1
	*/
	switch(mode->request.graphtype) {

        case KGIGT_TEXT16:
		CTRL1 = 0;
		break;

        case KGIGT_1BIT:
		CTRL1 = DAC_C1_1BPP;
		break;

	case KGIGT_4BIT:
		CTRL1 = DAC_C1_4BPP;
		break;

	case KGIGT_8BIT:
		CTRL1 = DAC_C1_8BPP;
		break;

	case KGIGT_15BIT:
		CTRL1 = DAC_C1_15BPP;
		break;

	case KGIGT_16BIT:
		CTRL1 = DAC_C1_16BPP;
		break;
	
	case KGIGT_32BIT:
		CTRL1 = DAC_C1_24BPP;
		break;

	default:
		INTERNAL_ERROR;	/* for safety, and to trick gcc */
	}
	DEBUG2("graphtype = %i", mode->request.graphtype);

	/*	You might want to turn CLUT bypass on for debugging only.
	*/
	#if 0
		DEBUG0(" .. CLUT bypass enabled.");
		private->ctrl1 |= DAC_C1_CLUT_BYPASS;
	#endif


	/*	setup ctrl2
	*/
	CTRL2 = ((mode->request.graphtype == KGIGT_TEXT16)
		? 0 : DAC_C2_ENHANCED_MODE)

		| ((mode->flags & FL_MON_INTERLACE)
			? DAC_C2_INTERLACED : 0)

		| PCLK;

	/*	setup ctrl_reg_3
	*/
	CTRL3 = CTRL3 & ~DAC_C3_LARGE_CURSOR;

	/*	set control registers (ctrl3 looks strange, but is correct ;-)
 	*/
	DEBUG2("ctrl0 %.2x, ctrl1 %.2x, ctrl2 %.2x, ctrl3 %.2x", \
		CTRL0, CTRL1, CTRL2, CTRL3);	
	DAC_out8(dpy, CTRL0 | DAC_C0_EXT_UNLOCK, DAC_CONTROL0);
	DAC_out8(dpy, DAC_PW_CRTL3_INDEX, DAC_PW_INDEX);
	DAC_out8(dpy, CTRL3, DAC_STATUSREG);

	DAC_out8(dpy, CTRL2, DAC_CONTROL2);
	DAC_out8(dpy, CTRL1, DAC_CONTROL1);
	DAC_out8(dpy, CTRL0, DAC_CONTROL0);
	DAC_out8(dpy, 0xFF, DAC_PIXEL_MASK);

	CUR_X = -1;
	CUR_Y = -1;
	DAC_out8(dpy, 0xFF, DAC_CURSOR_XL);
	DAC_out8(dpy, 0x0F, DAC_CURSOR_XH);
	DAC_out8(dpy, 0xFF, DAC_CURSOR_YL);
	DAC_out8(dpy, 0x0F, DAC_CURSOR_YH);

        dpy->set_clut = dac_set_clut;
	dpy->set_pointer_shape  = dac_set_pointer_shape;
        dpy->show_pointer = dac_show_pointer;
        dpy->hide_pointer = dac_hide_pointer;
	dpy->undo_pointer = NULL;
}

int kgim_ramdac_command (struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg)
{
	/*	this may set 
	**	-  dac_params (take care if 504 or 505 was detected).
	**	-  hardware cursor
	**	-  CLUT (or gamma correction table).
	*/

	kgi_clut clut;

	switch(cmd) {

	case RAMDAC_SETCLUT:
		clut.size = 256;
		clut.data=(kgi_color *)(graph->io_buf);
		dac_set_clut(graph->dev.dpy,&clut,0,256);
		return 0;
	case RAMDAC_GETCLUT:
		clut.size=256;
		clut.data=(kgi_color *)(graph->io_buf);
		graph->io_res=graph->io_buf;
		dac_get_clut(graph->dev.dpy,&clut,0,256);
		return 0;
	}

	/*	This driver can't do this. Maybe a simpler command will do.
	*/
	return -E(RAMDAC,NOSUP_ALWAYS_LOWER);	
}
